""":mod:`content_negotiation`
=============================

Provides better content-negotiation for flask.
"""
from flask import Response, request, abort

from renderers import TemplateRenderer
from decorators import provides
from media_type import acceptable_media_types, best_renderer, MediaType

__all__ = ('Render', 'MediaType', 'provides')


class Render(object):
    """Dynamic function class renders content.
    """
    def __init__(self, renderers=(TemplateRenderer(), )):
        super(Render, self).__init__()
        self.renderers = renderers

    def __call__(self, data, template=None, status=200, headers=None,
                 renderers=None, ctx=None):
        """Render `_data` to response.

        :param data: rendering target.
        :param template: path for jinja template '.html' can be ommited.
        :param status: status code for HTTP response.
        :param headers: additional header informations.
        :param renderers: list of renderer will be used.  :cosnt:`None` to use
            default renderers
        :param ctx: context for template renderer.  defualt is
            `{'data':data}`

        :returns: rendered response
        :rtype: :class:`flask.Response`

        You can use it like::

            from flask.ext.negotiation import render

            @app.route('/')
            def view():
                data = get_data()
                return render(data)

        Or::

            from flask.ext.negotiation import render

            @app.route('/user/<uid>')
            def user_view(uid):
                user = get_user(uid)
                return render(user, 'user/read')

        Or::

            from flask.ext.negotiation import render
            from flask.ext.negotiation import provides

            @app.route('/new_content/')
            @provides(TemplateRenderer, 'application/json')
            def content_view():
                content = create_content()
                return render(content, 'content/read.html', 201, ctx={
                    'content':content,
                    'author':content.author,
                })

        """
        renderers = renderers or self.renderers
        media_types = acceptable_media_types(request)
        renderer, rendered_media_type = best_renderer(renderers, media_types)
        if renderer is None:
            abort(406)
        body = renderer.render(data, template, ctx)
        return Response(body, status, headers, unicode(rendered_media_type),
                        content_type=unicode(rendered_media_type))



""":mod:`decorators` --- Decorators for Flask views
===================================================
"""
from functools import wraps

from flask import request
from werkzeug.exceptions import NotAcceptable

from renderers import Renderer
from media_type import acceptable_media_types, MediaType, choose_media_type


def provides(media_type, *args, **kwargs):
    """Decorator that recognizes acceptablility of view function.
    For example::

        from flask.ext.negotiation.decorators import provides

        @app.route('/json_only')
        @provides('application/json')
        def json_only():
            return json.dumps({'text':'JSON Mraz'})

    And you can provide multiple types like this::

        from flask.ext.negotiation.decorators import provides

        @app.route('/json_and_html')
        @provides('text/html', 'application/json')
        def json_and_html():
            data = get_data()
            return render(data)

    Or you can use renderer::

        from flask.ext.negotiation.renderes import template_renderer
        from flask.ext.negotiation.decorators import provides

        @app.route('/json_and_html')
        @provides(template_renderer, 'application/json')
        def json_and_html():
            data = get_data()
            return render(data)

    And you can handle choosen media type::

        from flask.ext.negotiation.decorators import provides
        @provides('application/json', 'text/html', to='provide_type')
        def handle_type(provide_type):
            return str(provide_type)

    `to` does *not* guarantee same media type with `render` function.
    """
    to = kwargs.get('to', None)
    # Collect media types
    media_types = []
    for media_type in (media_type, ) + args:
        if isinstance(media_type, MediaType):
            media_types.append(media_type)
        elif isinstance(media_type, type) and issubclass(media_type, Renderer):
            media_types += map(MediaType, media_type.__media_types__)
        elif isinstance(media_type, Renderer):
            media_types += media_type.media_types
        else:
            media_types.append(MediaType(media_type))

    # Decorator here
    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            acceptables = acceptable_media_types(request)
            acceptable = choose_media_type(acceptables, media_types)
            if acceptable is None:
                raise NotAcceptable()
            if not to is None:
                kwargs.update({to: acceptable})
            return fn(*args, **kwargs)
        return wrapper
    return decorator



""":mod:`media_type`
====================

HTTP media type
"""


def parse_header(s):
    """Parses parameter header
    """
    params = _parse_header_params(';'+s)
    key = params.pop(0).lower()
    pdict = {}
    for param in params:
        i = param.find('=')
        if i >= 0:
            name = param[:i].strip().lower()
            value = param[i+1:].strip()
            if len(value) >= 2 and value[0] == value[-1] == '"':
                value = value[1:-1]
                value = value.replace(r'\\', '\\').replace(r'\"', '"')
            pdict[name] = value
    return key, pdict


def _parse_header_params(s):
    li = []
    while s[:1] == ';':
        s = s[1:]
        end = s.find(';')
        while end > 0 and s.count('"', 0, end) % 2:
            # For quote
            end = s.find(';', end+1)
        end = end if end >= 0 else len(s)
        li.append(s[:end].strip())
        s = s[end:]
    return li


class MediaType(object):
    """Abstracted media type class.
    """
    def __init__(self, raw):
        raw = raw or ''
        self.raw = raw
        self.media_type, self.params = parse_header(raw)
        self.main_type, sep, self.sub_type = self.media_type.partition('/')

    def __contains__(self, other):
        for k, v in self.params.iteritems():
            if k != 'q' and other.params.get(k, None) != v:
                return False
        if self.main_type == '*' and self.sub_type == '*':
            return True
        if self.sub_type == '*' and self.main_type == other.main_type:
            return True
        if self.main_type == '*' and self.sub_type == other.sub_type:
            return True
        return self == other

    def __eq__(self, other):
        if isinstance(other, basestring):
            return unicode(self) == other
        return (self.main_type == other.main_type and
                self.sub_type == other.sub_type)

    def __repr__(self):
        return '<media type:' + str(self) + '>'

    def __str__(self):
        return unicode(self).encode('utf-8')

    def __unicode__(self):
        return u'; '.join([u'%s/%s' % (self.main_type, self.sub_type)] +
                          [u'%s=%s' % (k, v)
                           for k, v in self.params.iteritems()])

    def __cmp__(self, other):
        return cmp(float(self.quality),
                   float(other.quality))

    @property
    def quality(self):
        q = self.params.get('q', None)
        if q is None:
            return 1.0
        return float(q)


def acceptable_media_types(request):
    """Extract acceptable media types from request
    """
    if 'accept' in request.headers:
        li = [x.strip() for x in request.headers['accept'].split(',')]
    else:
        li = ['*/*']
    li = li or ['*/*']
    return sorted(map(MediaType, li), reverse=True)


def best_renderer(renderers, media_types):
    """Choose best renderer and media type
    """
    choosen_items = []
    for media_type in media_types:
        for renderer in renderers:
            choosen = renderer.choose_media_type(media_type)
            if not choosen is None:
                choosen_items.append((renderer, choosen, media_type))
    if not choosen_items:
        return None, None

    def cmp_types(first, second):
        renderer1, choosen1, media_type1 = first
        renderer2, choosen2, media_type2 = second
        if media_type1.quality == media_type2.quality:
            if media_type1 in media_types and media_type2 in media_types:
                cmp(media_types.index(media_type2),
                    media_types.index(media_type1))
            elif media_type1 in media_types:
                return media_type1
            elif media_type2 in media_types:
                return media_type2
            else:
                cmp(renderers.index(renderer2), renderers.index(renderer1))
        return cmp(media_type1.quality, media_type2.quality)

    return tuple(sorted(choosen_items, cmp=cmp_types)[-1][:2])


def choose_media_type(acceptables, media_types):
    """Choose best acceptable media type.
    :param acceptables: list of media type acceptable
    :param media_types: list of media type supported

    :returns: best acceptable media type or :const:`None` if cannot handle.
    """
    choosen = []
    for acceptable in acceptables:
        for media_type in media_types:
            if acceptable in media_type:
                choosen.append((acceptable, media_type))
    if not choosen:
        return None

    def cmp_types(first, second):
        acceptable1, media_type1 = first
        acceptable2, media_type2 = second
        if acceptable.quality == acceptable2.quality:
            return cmp(acceptables.index(acceptable2),
                       acceptables.index(acceptable1))
        return cmp(acceptable.quality, acceptable.quality)

    return sorted(choosen, cmp=cmp_types)[-1][0]


def can_accept(acceptables, media_types):
    """Determines acceptablility.
    :param acceptables: list of media type acceptable
    :param media_types: list of media type supported

    """
    return choose_media_type(acceptables, media_types) is not None



""":mod:`content_negotiation.renderers`
=======================================

Renderers
"""
import json
from abc import ABCMeta, abstractmethod
from flask import render_template
from functools import wraps
from media_type import MediaType


class Renderer(object):
    """Base renderer class.
    """
    __metaclass__ = ABCMeta

    __media_types__ = ()
    """A collection of supporting media-type :class:`string`s, subclasses must
    redefine this value.
    """

    @property
    def media_types(self):
        """Collections of abstracted media-types.
        """
        return (MediaType(x) for x in self.__media_types__)

    def can_render(self, media_type):
        """Determines that renderer can render `media_type`.
        """
        return not self.choose_media_type(media_type) is None

    def choose_media_type(self, media_type):
        """Chooses media type that will be rendered.
        """
        chosen_type = None
        for renderer_type in self.media_types:
            if media_type in renderer_type or renderer_type in media_type:
                chosen_type = renderer_type
        return chosen_type

    @abstractmethod
    def render(self, data, template=None, ctx=None):
        """Renders `data`.

        You must implement it
        """
        pass


class TemplateRenderer(Renderer):
    """Renders object to HTML response.
    """
    __media_types__ = ('text/html', )

    def __init__(self, ext='html'):
        super(TemplateRenderer, self).__init__()
        self.ext = ext

    def render(self, data, template=None, ctx=None):
        template = template or ''
        ext = '.' + self.ext
        if not template.endswith(ext):
            template += ext
        ctx = ctx or {
            'data': data
        }
        return render_template(template, **ctx)


class JSONRenderer(Renderer):
    """Renders object to json with JSONEncoder.
    """
    __media_types__ = ('application/json',)

    def __init__(self, encoder=json.JSONEncoder()):
        """:param encoder: encoder to be used with renderer.
        """
        super(JSONRenderer, self).__init__()
        self.encoder = encoder

    def render(self, data, template=None, ctx=None):
        return self.encoder.encode(data)


class FunctionRenderer(Renderer):
    """Renders object with a function.
    """
    def __init__(self, fn, media_types):
        super(FunctionRenderer, self).__init__()
        self.fn = fn
        self.__media_types__ = map(unicode, media_types)

    def render(self, data, template=None, ctx=None):
        return self.fn(data, template=template, ctx=ctx)

    def __call__(self, *args, **kwargs):
        return self.render(*args, **kwargs)


def renderer(*media_types):
    """Decorator that creates simple renderer with function.
    """
    def decorator(fn):
        renderer = wraps(fn)(FunctionRenderer(fn, media_types))
        return renderer
    return decorator

# default_renderers
template_renderer = TemplateRenderer()
json_renderer = JSONRenderer()



"""
Flask-Negotiation
==================

Provides better content negotiation for flask.
"""
import setuptools
from setuptools import setup

requires = [
    'Flask',
]

setup(name='Flask-Negotiation',
      version='0.1.8',
      url='http://blog.hardtack.me/',
      author='GunWoo Choi',
      author_email='6566gun@gmail.com',
      description='Better content negotiation for flask',
      long_description=__doc__,
      packages=setuptools.find_packages(exclude=('tests', )),
      include_package_data=True,
      zip_safe=False,
      platforms='any',
      install_requires=requires,
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
      ])
